@@ -5,7 +5,9 @@ produce an address space where that program is loaded and ready to run.
 The primary interface to CLE is the Loader class.
 """
 
-__version__ = "9.2.80"
+from __future__ import annotations
+
+__version__ = "9.2.182"
 
 from .address_translator import AT, AddressTranslator
 from .backends import (
@@ -35,6 +37,7 @@ from .backends import (
     Section,
     Segment,
     Soot,
+    SRec,
     StaticArchive,
     Symbol,
     SymbolSubType,
@@ -63,13 +66,15 @@ from .errors import (
     CLEError,
     CLEFileNotFoundError,
     CLEInvalidBinaryError,
+    CLEInvalidEncryptionError,
+    CLEInvalidFileFormatError,
     CLEMemoryError,
     CLEOperationError,
     CLEUnknownFormatError,
 )
 from .gdb import GDB_SEARCH_PATH, convert_info_proc_maps, convert_info_sharedlibrary
 from .loader import Loader
-from .memory import Clemory, ClemoryBase, ClemoryTranslator, ClemoryView
+from .memory import Clemory, ClemoryBase, ClemoryReadOnlyView, ClemoryTranslator, ClemoryView
 from .patched_stream import PatchedStream
 
 __all__ = [
@@ -88,6 +93,7 @@ __all__ = [
     "FunctionHint",
     "FunctionHintSource",
     "Hex",
+    "SRec",
     "Jar",
     "MachO",
     "MetaELF",
@@ -116,6 +122,8 @@ __all__ = [
     "CLEError",
     "CLEFileNotFoundError",
     "CLEInvalidBinaryError",
+    "CLEInvalidEncryptionError",
+    "CLEInvalidFileFormatError",
     "CLEOperationError",
     "CLEUnknownFormatError",
     "CLEMemoryError",
@@ -125,6 +133,7 @@ __all__ = [
     "Loader",
     "Clemory",
     "ClemoryBase",
+    "ClemoryReadOnlyView",
     "ClemoryView",
     "ClemoryTranslator",
     "PatchedStream",
