@@ -1,23 +1,30 @@
+# pylint:disable=bad-builtin
+from __future__ import annotations
+
 import copy
 import logging
 import os
+import pathlib
 import xml.etree.ElementTree
 from collections import OrderedDict, defaultdict
-from typing import Dict, List, Optional
+from typing import cast
 
 import archinfo
-import elftools
 from elftools.common.exceptions import DWARFError, ELFError, ELFParseError
 from elftools.dwarf import callframe
+from elftools.dwarf.compileunit import CompileUnit
 from elftools.dwarf.descriptions import describe_attr_value, describe_form_class
 from elftools.dwarf.die import DIE
 from elftools.dwarf.dwarf_expr import DWARFExprParser
 from elftools.dwarf.dwarfinfo import DWARFInfo
+from elftools.dwarf.ranges import BaseAddressEntry, RangeEntry
 from elftools.elf import dynamic, elffile, enums, sections
+from elftools.elf.relocation import RelocationSection, RelrRelocationSection
 from sortedcontainers import SortedDict
 
 from cle.address_translator import AT
 from cle.backends.backend import ExceptionHandling, FunctionHint, FunctionHintSource, register_backend
+from cle.backends.inlined_function import InlinedFunction
 from cle.errors import CLECompatibilityError, CLEError, CLEInvalidBinaryError
 from cle.patched_stream import PatchedStream
 from cle.utils import ALIGN_DOWN, ALIGN_UP, get_mmaped_data, stream_or_path
@@ -26,9 +33,9 @@ from .compilation_unit import CompilationUnit
 from .hashtable import ELFHashTable, GNUHashTable
 from .lsda import LSDAExceptionTable
 from .metaelf import MetaELF, maybedecode
-from .regions import ELFSection, ELFSegment
+from .regions import ELFSection, ELFSegment, Section
 from .relocation import get_relocation
-from .relocation.generic import MipsGlobalReloc, MipsLocalReloc
+from .relocation.generic import GenericRelativeReloc, MipsGlobalReloc, MipsLocalReloc
 from .subprogram import LexicalBlock, Subprogram
 from .symbol import ELFSymbol, Symbol, SymbolType
 from .variable import Variable
@@ -44,10 +51,15 @@ log = logging.getLogger(name=__name__)
 
 __all__ = ("ELFSymbol", "ELF")
 
+_NON_ALLOCATED_SECTION_NAMES = {  # Sections that do not occupy memory at runtime
+    "SHT_NOTE",
+    "SHT_MIPS_REGINFO",
+}
+
 
 # map 'e_machine' ELF header values (represented as `short int`s) to human-readable format (string)
 # There are mappings missing currently in `elftools`, so we provide them ourselves
-additional_e_machine_mappings: Dict[int, str] = {
+additional_e_machine_mappings: dict[int, str] = {
     247: "Linux BPF - in-kernel virtual machine",
     252: "C-SKY",
     0x5441: "Fujitsu FR-V",
@@ -79,6 +91,12 @@ class ELF(MetaELF):
         **kwargs,
     ):
         super().__init__(*args, **kwargs)
+        self.set_load_args(
+            addend=addend,
+            debug_symbols=debug_symbols,
+            discard_section_headers=discard_section_headers,
+            discard_program_headers=discard_program_headers,
+        )
         patch_undo = []
         try:
             self._reader = elffile.ELFFile(self._binary_stream)
@@ -103,11 +121,11 @@ class ELF(MetaELF):
 
             try:
                 self._reader = elffile.ELFFile(self._binary_stream)
-            except Exception as e:  # pylint: disable=broad-except
-                raise CLECompatibilityError from e
+            except Exception as e1:  # pylint: disable=broad-except
+                raise CLECompatibilityError from e1
 
         # Get an appropriate archinfo.Arch for this binary, unless the user specified one
-        if self.arch is None:
+        if self._arch is None:
             self.set_arch(self.extract_arch(self._reader))
         else:
             try:
@@ -148,9 +166,10 @@ class ELF(MetaELF):
         # DWARF data
         self.has_dwarf_info = bool(self._reader.has_dwarf_info())
         self.build_id = None
-        self.addr_to_line = SortedDict()
-        self.variables: Optional[List[Variable]] = None
-        self.compilation_units: Optional[List[CompilationUnit]] = None
+        self.addr_to_line: SortedDict[int, set[tuple[int, int]]] = SortedDict()
+        self.variables: list[Variable] | None = None
+        self.compilation_units: list[CompilationUnit] | None = None
+        self.functions_debug_info: dict[int, Subprogram] = {}
 
         # misc
         self._entry = self._reader.header.e_entry
@@ -258,7 +277,7 @@ class ELF(MetaELF):
         identstring = stream.read(0x1000)
         stream.seek(0)
         if identstring.startswith(b"\x7fELF"):
-            if elftools.elf.elffile.ELFFile(stream).header["e_type"] == "ET_CORE":
+            if elffile.ELFFile(stream).header["e_type"] == "ET_CORE":
                 return False
             return True
         return False
@@ -276,14 +295,14 @@ class ELF(MetaELF):
             return None  # No attrs here!
         attrs_sub_sec = None
         for subsec in attrs_sec.subsections:
-            if isinstance(subsec, elftools.elf.sections.ARMAttributesSubsection):
+            if isinstance(subsec, sections.ARMAttributesSubsection):
                 attrs_sub_sec = subsec
                 break
         if not attrs_sub_sec:
             return None  # None here either
         attrs_sub_sub_sec = None
         for subsubsec in attrs_sub_sec.subsubsections:
-            if isinstance(subsubsec, elftools.elf.sections.ARMAttributesSubsubsection):
+            if isinstance(subsubsec, sections.ARMAttributesSubsubsection):
                 attrs_sub_sub_sec = subsubsec
                 break
         if not attrs_sub_sub_sec:
@@ -305,7 +324,7 @@ class ELF(MetaELF):
                 arch_str = additional_e_machine_mappings[e_machine_header_val]
             else:
                 raise CLECompatibilityError(
-                    "The `e_machine` header value " "of the ELF file is not known: %d" % e_machine_header_val
+                    f"The `e_machine` header value of the ELF file is not known: {e_machine_header_val}"
                 )
         else:
             assert False  # assumption that `reader['e_machine'] returns `str` or `int` is violated
@@ -374,14 +393,14 @@ class ELF(MetaELF):
     def symbols_by_name(self):
         return self._symbols_by_name.copy()
 
-    def get_symbol(self, symid, symbol_table=None):  # pylint: disable=arguments-differ
+    def get_symbol(self, symid, symbol_table=None):  # pylint: disable=arguments-differ,arguments-renamed
         """
         Gets a Symbol object for the specified symbol.
 
         :param symid: Either an index into .dynsym or the name of a symbol.
         """
         version = None
-        if type(symid) is int:
+        if isinstance(symid, int):
             if symid == 0:
                 # special case the null symbol, this is important for static binaries
                 return self._nullsymbol
@@ -398,7 +417,7 @@ class ELF(MetaELF):
                 return cached
             if self.hashtable is not None and symbol_table is self.hashtable.symtab and self._vertable is not None:
                 version = self._vertable.get_symbol(symid).entry.ndx
-        elif type(symid) is str:
+        elif isinstance(symid, str):
             if not symid:
                 log.warning("Trying to resolve a symbol by its empty name")
                 return None
@@ -434,7 +453,7 @@ class ELF(MetaELF):
         delta = new_base - self.linked_base
         super().rebase(new_base)
 
-        self.addr_to_line = {addr + delta: value for addr, value in self.addr_to_line.items()}
+        self.addr_to_line = SortedDict((addr + delta, value) for addr, value in self.addr_to_line.items())
 
     #
     # Private Methods
@@ -512,9 +531,7 @@ class ELF(MetaELF):
                 self.binary,
             )
         if ph.p_filesz > ph.p_memsz:
-            raise CLEInvalidBinaryError(
-                "ELF file %s is loading a segment with an inappropriate allocation" % self.binary
-            )
+            raise CLEInvalidBinaryError(f"ELF file {self.binary} is loading a segment with an inappropriate allocation")
 
         mapstart = ALIGN_DOWN(ph.p_vaddr, self.loader.page_size)
         mapend = ALIGN_UP(ph.p_vaddr + ph.p_filesz, self.loader.page_size)
@@ -549,15 +566,24 @@ class ELF(MetaELF):
 
         self.memory.add_backer(AT.from_lva(mapstart, self).to_rva(), data, overwrite=True)
 
-    def _make_reloc(self, readelf_reloc, symbol, dest_section=None):
+    def _make_reloc(self, readelf_reloc, symbol, dest_section: Section | None = None, is_relr: bool = False):
         addend = readelf_reloc.entry.r_addend if readelf_reloc.is_RELA() else None
-        RelocClass = get_relocation(self.arch.name, readelf_reloc.entry.r_info_type)
+        # RELR relocations are compressed forms of R_*_RELATIVE relocations.
+        # If is_relr is True, the readelf_reloc.entry object does not have r_info_type,
+        # so we cannot dispatch through get_relocation. Instead, handle it as a special
+        # case by always using GenericRelativeReloc.
+        RelocClass = (
+            GenericRelativeReloc if is_relr else get_relocation(self.arch.name, readelf_reloc.entry.r_info_type)
+        )
         if RelocClass is None:
             return None
 
         address = AT.from_lva(readelf_reloc.entry.r_offset, self).to_rva()
         if dest_section is not None:
-            address += dest_section.remap_offset
+            # note to the intrepid explorer: this code was changed as per cle#467
+            # address += dest_section.remap_offset
+            if self.is_relocatable:
+                address += AT.from_mva(dest_section.vaddr, self).to_rva()
 
         try:
             return RelocClass(self, symbol, address, addend)
@@ -576,7 +602,7 @@ class ELF(MetaELF):
 
         try:
             for entry in dwarf.EH_CFI_entries():
-                if type(entry) is callframe.FDE:
+                if isinstance(entry, callframe.FDE):
                     self.function_hints.append(
                         FunctionHint(
                             entry.header["initial_location"],
@@ -599,7 +625,11 @@ class ELF(MetaELF):
         try:
             lsda = LSDAExceptionTable(self._binary_stream, self.arch.bits, self.arch.memory_endness == "Iend_LE")
             for entry in dwarf.EH_CFI_entries():
-                if type(entry) is callframe.FDE and hasattr(entry, "lsda_pointer") and entry.lsda_pointer is not None:
+                if (
+                    isinstance(entry, callframe.FDE)
+                    and hasattr(entry, "lsda_pointer")
+                    and entry.lsda_pointer is not None
+                ):
                     # function address
                     func_addr = entry.header.get("initial_location", None)
                     if func_addr is None:
@@ -653,20 +683,35 @@ class ELF(MetaELF):
                 else:
                     file_entry = lineprog.header["file_entry"][line.state.file - 1]
                     if file_entry["dir_index"] == 0:
-                        filename = os.path.join(comp_dir, file_entry.name.decode())
+                        filename = pathlib.PurePosixPath(comp_dir) / file_entry.name.decode()
                     else:
-                        filename = os.path.join(
-                            comp_dir,
-                            lineprog.header["include_directory"][file_entry["dir_index"] - 1].decode(),
-                            file_entry.name.decode(),
+                        filename = (
+                            pathlib.PurePosixPath(comp_dir)
+                            / lineprog.header["include_directory"][file_entry["dir_index"] - 1].decode()
+                            / file_entry.name.decode()
                         )
                     file_cache[line.state.file] = filename
 
                 relocated_addr = AT.from_lva(line.state.address, self).to_mva()
-                self.addr_to_line[relocated_addr] = (filename, line.state.line)
+                if relocated_addr not in self.addr_to_line:
+                    self.addr_to_line[relocated_addr] = set()
+                self.addr_to_line[relocated_addr].add((str(filename), line.state.line))
 
     @staticmethod
-    def _load_low_high_pc_form_die(die: DIE):
+    def _load_ranges_form_die(die: DIE, aranges) -> list[tuple[int, int]] | None:
+        if aranges is not None and "DW_AT_ranges" in die.attributes:
+            base_addr = 0
+            result = []
+            for entry in aranges.get_range_list_at_offset(die.attributes["DW_AT_ranges"].value, die.cu):
+                if isinstance(entry, BaseAddressEntry):
+                    base_addr = entry.base_address
+                elif isinstance(entry, RangeEntry):
+                    result.append((base_addr + entry.begin_offset, base_addr + entry.end_offset))
+            return result
+        return None
+
+    @staticmethod
+    def _load_low_high_pc_form_die(die: DIE) -> tuple[int | None, int | None]:
         """
         Load low and high pc from a DIE.
 
@@ -703,8 +748,9 @@ class ELF(MetaELF):
         :param dwarf:   The DWARF info object from pyelftools.
         :return:        None
         """
-        compilation_units: List[CompilationUnit] = []
-        type_list: Dict[int, VariableType] = {}
+        compilation_units: list[CompilationUnit] = []
+        type_list: dict[int, VariableType] = {}
+        aranges = dwarf.range_lists()
 
         for cu in dwarf.iter_CUs():
             expr_parser = DWARFExprParser(cu.structs)
@@ -730,12 +776,17 @@ class ELF(MetaELF):
             die_comp_dir = top_die.attributes.get("DW_AT_comp_dir", None)
             die_low_pc, die_high_pc = self._load_low_high_pc_form_die(top_die)
             die_lang = top_die.attributes.get("DW_AT_language", None)
+            die_ranges = self._load_ranges_form_die(top_die, aranges)
 
+            if die_low_pc is None and die_ranges is not None:
+                die_low_pc = min(x for x, _ in die_ranges)
+
+            # pylint: disable=too-many-boolean-expressions
             if (
                 die_name is None
                 or die_comp_dir is None
                 or die_low_pc is None
-                or die_high_pc is None
+                or (die_high_pc is None and die_ranges is None)
                 or die_lang is None
             ):
                 continue
@@ -744,38 +795,77 @@ class ELF(MetaELF):
             die_comp_dir = die_comp_dir.value.decode("utf-8")
             die_lang = describe_attr_value(die_lang, top_die, top_die.offset)
 
-            cu_ = CompilationUnit(die_name, die_comp_dir, die_low_pc, die_high_pc, die_lang, self)
+            cu_ = CompilationUnit(die_name, die_comp_dir, die_low_pc, die_high_pc, die_lang, self, die_ranges)
             compilation_units.append(cu_)
-
-            for die_child in cu.iter_DIE_children(top_die):
-                if die_child.tag == "DW_TAG_variable":
-                    # load global variable
-                    var = Variable.from_die(die_child, expr_parser, self)
-                    var.decl_file = cu_.file_path
-                    cu_.global_variables.append(var)
-                elif die_child.tag == "DW_TAG_subprogram":
-                    # load subprogram
-                    sub_prog = self._load_die_lex_block(die_child, expr_parser, type_list, cu, cu_.file_path, None)
-                    if sub_prog is not None:
-                        cu_.functions[sub_prog.low_pc] = sub_prog
+            self._load_die_namespace(top_die, dwarf, aranges, cu_, expr_parser, type_list, cu, [])
 
         self.type_list = type_list
         self.compilation_units = compilation_units
 
-    def _load_die_lex_block(self, die: DIE, expr_parser, type_list, cu, file_path, subprogram) -> LexicalBlock:
+    def _load_die_namespace(
+        self,
+        die: DIE,
+        dwarf: DWARFInfo,
+        aranges,
+        cu_: CompilationUnit,
+        expr_parser,
+        type_list,
+        cu,
+        namespace: list[str],
+    ) -> None:
+        for die_child in die.iter_children():
+            if die_child.tag == "DW_TAG_variable":
+                # load global variable
+                var = Variable.from_die(die_child, expr_parser, self, namespace=namespace)
+                var.decl_file = cu_.file_path
+                cu_.global_variables.append(var)
+            elif die_child.tag == "DW_TAG_subprogram":
+                # load subprogram
+                sub_prog = self._load_die_lex_block(
+                    die_child, dwarf, aranges, expr_parser, type_list, cu, cu_.file_path, None, namespace=namespace
+                )
+                if sub_prog is not None:
+                    assert isinstance(sub_prog, Subprogram)
+                    cu_.functions[sub_prog.low_pc] = sub_prog
+                    self.functions_debug_info[sub_prog.low_pc] = sub_prog
+            elif die_child.tag == "DW_TAG_namespace":
+                if "DW_AT_name" in die_child.attributes:
+                    new_namespace = namespace + [die_child.attributes["DW_AT_name"].value.decode("utf-8")]
+                else:
+                    new_namespace = namespace
+                self._load_die_namespace(die_child, dwarf, aranges, cu_, expr_parser, type_list, cu, new_namespace)
+
+    def _load_die_lex_block(
+        self,
+        die: DIE,
+        dwarf: DWARFInfo,
+        aranges,
+        expr_parser,
+        type_list,
+        cu: CompileUnit,
+        file_path,
+        subprogram,
+        namespace: list[str] | None = None,
+    ) -> LexicalBlock | None:
         if "DW_AT_name" in die.attributes:
-            name = die.attributes["DW_AT_name"].value.decode("utf-8")
+            name = "::".join((namespace or []) + [die.attributes["DW_AT_name"].value.decode("utf-8")])
+        elif "DW_AT_abstract_origin" in die.attributes:
+            origin = die.get_DIE_from_attribute("DW_AT_abstract_origin")
+            name = self._dwarf_get_name_with_namespace(origin)
         else:
             name = None
 
         low_pc, high_pc = self._load_low_high_pc_form_die(die)
+        ranges = None
         if low_pc is None or high_pc is None:
-            return None
+            ranges = self._load_ranges_form_die(die, aranges)
+            if ranges is None:
+                return None
 
         if subprogram is None:
-            subprogram = block = Subprogram(name, low_pc, high_pc)
+            subprogram = block = Subprogram(name, low_pc, high_pc, ranges)
         else:
-            block = LexicalBlock(low_pc, high_pc)
+            block = LexicalBlock(low_pc, high_pc, ranges)
 
         for sub_die in cu.iter_DIE_children(die):
             if sub_die.tag in ["DW_TAG_variable", "DW_TAG_formal_parameter"]:
@@ -784,12 +874,43 @@ class ELF(MetaELF):
                 var.decl_file = file_path
                 subprogram.local_variables.append(var)
             elif sub_die.tag == "DW_TAG_lexical_block":
-                sub_block = self._load_die_lex_block(sub_die, expr_parser, type_list, cu, file_path, subprogram)
+                sub_block = self._load_die_lex_block(
+                    sub_die, dwarf, aranges, expr_parser, type_list, cu, file_path, subprogram, namespace
+                )
                 if sub_block is not None:
                     block.child_blocks.append(sub_block)
+            elif sub_die.tag == "DW_TAG_inlined_subroutine":
+                subr = InlinedFunction()
+                low_pc, high_pc = self._load_low_high_pc_form_die(sub_die)
+                if low_pc is not None and high_pc is not None:
+                    subr.ranges.append((low_pc, high_pc))
+                elif "DW_AT_ranges" in sub_die.attributes:
+                    aranges = dwarf.range_lists()
+                    ranges = self._load_ranges_form_die(sub_die, aranges)
+                    if ranges is not None:
+                        subr.ranges = ranges
+                if "DW_AT_abstract_origin" in sub_die.attributes:
+                    origin = sub_die.get_DIE_from_attribute("DW_AT_abstract_origin")
+                    subr.name = self._dwarf_get_name_with_namespace(origin)
+                subprogram.inlined_functions.append(subr)
 
         return block
 
+    @staticmethod
+    def _dwarf_get_name_with_namespace(origin_die: DIE) -> str | None:
+        name_pieces = []
+        while True:
+            name_piece = origin_die.attributes.get("DW_AT_name", None)
+            if name_piece is not None:
+                name_pieces.append(name_piece.value.decode("utf-8"))
+
+            origin_die = cast(DIE, origin_die.get_parent())
+            if origin_die is None or origin_die.tag != "DW_TAG_namespace":
+                break
+        if name_pieces:
+            return "::".join(reversed(name_pieces))
+        return None
+
     #
     # Private Methods... really. Calling these out of context
     # will probably break things. Caveat emptor.
@@ -988,7 +1109,7 @@ class ELF(MetaELF):
                     "sh_flags": 0,
                     "sh_addralign": 0,
                 }
-                readelf_relocsec = elffile.RelocationSection(fakerelheader, "reloc_cle", self._reader)
+                readelf_relocsec = RelocationSection(fakerelheader, "reloc_cle", self._reader)
                 # support multiple versions of pyelftools
                 readelf_relocsec.stream = self.memory
                 readelf_relocsec._stream = self.memory
@@ -1009,13 +1130,33 @@ class ELF(MetaELF):
                     "sh_flags": 0,
                     "sh_addralign": 0,
                 }
-                readelf_jmprelsec = elffile.RelocationSection(fakejmprelheader, "jmprel_cle", self._reader)
+                readelf_jmprelsec = RelocationSection(fakejmprelheader, "jmprel_cle", self._reader)
                 # support multiple versions of pyelftools
                 readelf_jmprelsec.stream = self.memory
                 readelf_jmprelsec._stream = self.memory
                 readelf_jmprelsec.elffile = None
                 self.__register_relocs(readelf_jmprelsec, dynsym, force_jmprel=True)
 
+            # try to parse relocations out of a table of type DT_RELR
+            if "DT_RELR" in self._dynamic:
+                reloffset = AT.from_lva(self._dynamic["DT_RELR"], self).to_rva()
+                if "DT_RELRSZ" not in self._dynamic:
+                    raise CLEInvalidBinaryError("Dynamic section contains DT_RELR but not DT_RELRSZ")
+                relsz = self._dynamic["DT_RELRSZ"]
+                fakerelheader = {
+                    "sh_offset": reloffset,
+                    "sh_type": "SHT_" + "RELR",
+                    "sh_entsize": self._reader.structs.Elf_Relr.sizeof(),
+                    "sh_size": relsz,
+                    "sh_flags": 0,
+                    "sh_addralign": 0,
+                }
+                readelf_relrsec = RelrRelocationSection(fakerelheader, "relr_cle", self._reader)
+                # support multiple versions of pyelftools
+                readelf_relrsec.stream = self.memory
+                readelf_relrsec.elffile = None
+                self.__register_relocs(readelf_relrsec, dynsym)
+
             self.__register_section_symbols(dynsym)
 
     def __parse_rpath(self, runpath, rpath):
@@ -1042,9 +1183,6 @@ class ELF(MetaELF):
         return parts
 
     def __register_relocs(self, section, dynsym=None, force_jmprel=False):
-        if not self.loader._perform_relocations:
-            return
-
         got_min = got_max = 0
 
         if not force_jmprel:
@@ -1071,14 +1209,22 @@ class ELF(MetaELF):
                     return
 
         symtab = self._reader.get_section(section.header["sh_link"]) if "sh_link" in section.header else dynsym
-        if isinstance(symtab, elftools.elf.sections.NullSection):
+        if isinstance(symtab, sections.NullSection):
             # Oh my god Atmel please stop
             symtab = self._reader.get_section_by_name(".symtab")
         relocs = []
         for readelf_reloc in section.iter_relocations():
+            # Handle packed RELR relocations specially
+            # These entries lack r_info_sym/r_info_type, so this branch must be
+            # exclusive — do NOT fall through into later cases.
+            if section.header.get("sh_type") == "SHT_RELR":
+                reloc = self._make_reloc(readelf_reloc, self._nullsymbol, dest_sec, True)
+                if reloc is not None:
+                    relocs.append(reloc)
+                    self.relocs.append(reloc)
             # MIPS64 is just plain old fucked up
             # https://www.sourceware.org/ml/libc-alpha/2003-03/msg00153.html
-            if self.arch.name == "MIPS64":
+            elif self.arch.name == "MIPS64":
                 if not hasattr(readelf_reloc.entry, "r_info_type2") and hasattr(readelf_reloc.entry, "r_info_type3"):
                     raise CLECompatibilityError(
                         "This code relies on `pyelftools` features that are not available on versions 0.26 and below."
@@ -1199,9 +1345,10 @@ class ELF(MetaELF):
                 align = sec_readelf.header["sh_addralign"]
                 if align > 0:
                     new_addr = (new_addr + (align - 1)) // align * align
-
                 remap_offset = new_addr - sh_addr
-                new_addr += sec_readelf.header["sh_size"]  # address for next section
+
+                if sec_readelf.header["sh_type"] not in _NON_ALLOCATED_SECTION_NAMES:
+                    new_addr += sec_readelf.header["sh_size"]
 
             section = ELFSection(sec_readelf, remap_offset=remap_offset)
             sec_list.append((sec_readelf, section))
@@ -1211,10 +1358,13 @@ class ELF(MetaELF):
             self.sections_map[section.name] = section
 
         for sec_readelf, section in sec_list:
-            if isinstance(sec_readelf, elffile.SymbolTableSection):
+            if isinstance(sec_readelf, sections.SymbolTableSection):
                 self.__register_section_symbols(sec_readelf)
-            if isinstance(sec_readelf, elffile.RelocationSection) and not (
-                "DT_REL" in self._dynamic or "DT_RELA" in self._dynamic or "DT_JMPREL" in self._dynamic
+            if isinstance(sec_readelf, RelocationSection | RelrRelocationSection) and not (
+                "DT_REL" in self._dynamic
+                or "DT_RELA" in self._dynamic
+                or "DT_JMPREL" in self._dynamic
+                or "DT_RELR" in self._dynamic
             ):
                 self.__register_relocs(sec_readelf, dynsym=None)
 
@@ -1228,7 +1378,7 @@ class ELF(MetaELF):
                             b"\0" * sec_readelf.header["sh_size"],
                             overwrite=True,
                         )
-                    elif section.type == "SHT_NOTE":
+                    elif section.type in _NON_ALLOCATED_SECTION_NAMES:
                         pass  # observed this case in angr/angr#3829
                     else:  # elif section.type == 'SHT_PROGBITS':
                         self.memory.add_backer(
@@ -1296,10 +1446,10 @@ class ELF(MetaELF):
         return True
 
     def __neuter_streams(self, obj):
-        if isinstance(obj, elftools.elf.dynamic._DynamicStringTable):
+        if isinstance(obj, dynamic._DynamicStringTable):
             obj._stream = self.memory
             obj._table_offset = self._offset_to_rva(obj._table_offset)
-        elif isinstance(obj, elftools.elf.sections.Section):
+        elif isinstance(obj, sections.Section):
             if obj.header.sh_type == "SHT_NOBITS":
                 obj.stream = None
                 obj.elffile = None
@@ -1309,7 +1459,7 @@ class ELF(MetaELF):
                 obj.elffile = None
                 obj.header.sh_offset = self._offset_to_rva(obj.header.sh_offset)
         else:
-            raise TypeError("Can't convert %r" % type(obj))
+            raise TypeError(f"Can't convert {type(obj)!r}")
 
     def _offset_to_rva(self, offset):
         return AT.from_mva(self.offset_to_addr(offset), self).to_rva()
@@ -1323,7 +1473,7 @@ class ELF(MetaELF):
                 return
 
             for sec_readelf in elf.iter_sections():
-                if isinstance(sec_readelf, elffile.SymbolTableSection):
+                if isinstance(sec_readelf, sections.SymbolTableSection):
                     self.__register_section_symbols(sec_readelf)
                 elif sec_readelf.header.sh_type == "SHT_NOTE":
                     self.__register_notes(sec_readelf)
