@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import logging
 from enum import Enum
 from typing import TYPE_CHECKING
@@ -27,6 +29,7 @@ class SymbolType(Enum):
     TYPE_OBJECT = 3
     TYPE_SECTION = 4
     TYPE_TLS_OBJECT = 5
+    TYPE_FUNCTION_OR_OBJECT = 6
 
 
 class SymbolSubType(Enum):
@@ -60,11 +63,11 @@ class Symbol:
     :ivar str resolvewith:  The name of the library we must use to resolve this symbol, or None if none is required.
     """
 
-    def __init__(self, owner: "Backend", name: str, relative_addr: int, size: int, sym_type: SymbolType):
+    def __init__(self, owner: Backend, name: str, relative_addr: int, size: int, sym_type: SymbolType):
         """
         Not documenting this since if you try calling it, you're wrong.
         """
-        self.owner = owner
+        self.owner: Backend = owner
         self.name = name
         self.relative_addr = relative_addr
         self.size = size
@@ -113,7 +116,7 @@ class Symbol:
         """
         Whether this symbol is a function
         """
-        return self.type is SymbolType.TYPE_FUNCTION
+        return self.type in {SymbolType.TYPE_FUNCTION, SymbolType.TYPE_FUNCTION_OR_OBJECT}
 
     # These may be overridden in subclasses
     is_static = False
